#include "matrix.h"

#include <string.h>

// ----------------------------------------------------------------------------
// 便利関数
// ----------------------------------------------------------------------------

bool is_same_size(matrix mat1, matrix mat2) {
    return mat1.rows == mat2.rows && mat1.cols == mat2.cols;
}

bool is_valid(matrix mat) {
    return mat.rows > 0 && mat.cols > 0 && mat.elems != NULL;
}

double add(double a, double b) { return a + b; }
double sub(double a, double b) { return a - b; }
double mul(double a, double b) { return a * b; }

// ----------------------------------------------------------------------------
// 行列演算用関数群
// ----------------------------------------------------------------------------

// mat_alloc:行列要素用のメモリを確保する
bool mat_alloc(matrix *mat, int rows, int cols)
{
    if (rows <= 0 || cols <= 0) {
        return false;
    }

    mat->elems = (double*)calloc(rows * cols, sizeof(double));
    if (!mat->elems) {
        return false;
    }

    mat->rows = rows;
    mat->cols = cols;
    return true;
}

// mat_free:使い終わった行列のメモリを解放する
void mat_free(matrix *mat)
{
    if (mat->elems != NULL) {
        free(mat->elems);
        mat->rows = 0;
        mat->cols = 0;
        mat->elems = NULL;
    }
}

// mat_print:行列の中身を表示する
void mat_print(matrix mat)
{
    if (!is_valid(mat)) {
        fprintf(stderr, "Matrix is NULL or zero size!\n");
        return;
    }

    for (int i = 0; i < mat.rows; i++) {
        for (int j = 0; j < mat.cols; j++) {
            printf("%6.4f%s", mat_elem(mat, i, j), (j == mat.cols - 1) ? "\n" : "  ");
        }
    }
}

// mat_copy:srcの中身を*dstにコピーする
bool mat_copy(matrix *dst, matrix src)
{
    if (!is_same_size(src, *dst)) return false;
    if (!is_valid(src) || !is_valid(*dst)) return false;

    memcpy(dst->elems, src.elems, sizeof(double) * src.rows * src.cols);
    return true;
}

// 二項演算子を定義するための汎用関数
bool mat_binary_op(matrix *res, matrix mat1, matrix mat2, double (*op)(double, double)) {
    if (!is_valid(*res) || !is_valid(mat1) || !is_valid(mat2)) return false;
    if (!is_same_size(mat1, mat2) || !is_same_size(*res, mat1)) return false;

    for (int i = 0; i < mat1.rows; i++) {
        for (int j = 0; j < mat1.cols; j++) {
            const double v1 = mat_elem(mat1, i, j);
            const double v2 = mat_elem(mat2, i, j);
            mat_elem(*res, i, j) = op(v1, v2);
        }
    }

    return true;
}

// mat_add:mat1+mat2を*resに代入する
bool mat_add(matrix *res, matrix mat1, matrix mat2)
{
    return mat_binary_op(res, mat1, mat2, add);
}

// mat_sub:mat1-mat2を*resに代入する
bool mat_sub(matrix *res, matrix mat1, matrix mat2)
{
    return mat_binary_op(res, mat1, mat2, sub);
}

// mat_mul:mat1とmat2の行列積を*resに代入する
bool mat_mul(matrix *res, matrix mat1, matrix mat2)
{
    if (!is_valid(*res) || !is_valid(mat1) || !is_valid(mat2)) return false;
    if (res->rows != mat1.rows || mat1.cols != mat2.rows || res->cols != mat2.cols) return false;

    matrix temp;
    if (!mat_alloc(&temp, res->rows, res->cols)) return false;

    for (int i = 0; i < res->rows; i++) {
        for (int j = 0; j < res->cols; j++) {
            mat_elem(temp, i, j) = 0.0;
            for (int k = 0; k < mat1.cols; k++) {
                mat_elem(temp, i, j) += mat_elem(mat1, i, k) * mat_elem(mat2, k, j);
            }
        }
    }
    
    mat_copy(res, temp);
    mat_free(&temp);
    
    return true;
}


// mat_muls:matをc倍（スカラー倍）した結果を*resに代入する
bool mat_muls(matrix *res, matrix mat, double c)
{
    if (!is_valid(*res) || !is_valid(mat)) return false;
    if (!is_same_size(*res, mat)) return false;

    for (int i = 0; i < mat.rows; i++) {
        for (int j = 0; j < mat.cols; j++) {
            mat_elem(*res, i, j) = mat_elem(mat, i, j) * c;
        }
    }
    return true;
}

// mat_trans:matの転置行列を*resに代入する
bool mat_trans(matrix *res, matrix mat)
{
    if (res->rows != mat.cols || res->cols != mat.rows) return false;

    matrix temp;
    mat_alloc(&temp, res->rows, res->cols);

    for (int i = 0; i < res->rows; i++) {
        for (int j = 0; j < res->cols; j++) {
            mat_elem(temp, i, j) = mat_elem(mat, j, i);
        }
    }

    mat_copy(res, temp);
    mat_free(&temp);

    return true;
}

// mat_unit:単位行列を与える
bool mat_ident(matrix *mat)
{
    if (mat->rows != mat->cols) return false;
    for (int i = 0; i < mat->rows; i++) {
        for (int j = 0; j < mat->cols; j++) {
            mat_elem(*mat, i, j) = (i == j) ? 1.0 : 0.0;
        }
    }
    return true;
}

// mat_equal:mat1とmat2が等しければtrueを返す
bool mat_equal(matrix mat1, matrix mat2)
{
    static const double epsilon = 1.0e-12;

    if (!is_same_size(mat1, mat2)) return false;
    
    for (int i = 0; i < mat1.rows; i++) {
        for (int j = 0; j < mat1.cols; j++) {
            const double v1 = mat_elem(mat1, i, j);
            const double v2 = mat_elem(mat2, i, j);
            if (fabs(v1 - v2) >= epsilon) return false; 
        }
    }
    return true;
}

// mat_solve:連立一次方程式 ax=b を解く．ピボット選択付き
bool mat_solve(matrix *x, matrix A_, matrix b_)
{
    static const double epsilon = 1.0e-12;

    if (!is_valid(*x) || !is_valid(A_) || !is_valid(b_)) return false;
    if (A_.cols != x->rows || A_.rows != b_.rows || x->cols != b_.cols) return false;

    // 行列Aと行列bの値を書き換えないよう，別の行列を用意する
    matrix A, b;
    if (!mat_alloc(&A, A_.rows, A_.cols)) return false;
    if (!mat_alloc(&b, b_.rows, b_.cols)) return false;
    mat_copy(&A, A_);
    mat_copy(&b, b_);

    // ガウスの消去法
    int pivot = 0;
    for (int i = 0; i < A.rows - 1; i++) {
        // ピボット選択
        pivot = i;
        for (int j = i + 1; j < A.rows; j++) {
            if (fabs(mat_elem(A, pivot, i)) < fabs(mat_elem(A, j, i))) {
                pivot = j;
            }
        }

        // 行の入れ替え
        for (int j = i; j < A.cols; j++) {
            swap(mat_elem(A, i, j), mat_elem(A, pivot, j));
        }

        for (int j = 0; j < b.cols; j++) {
            swap(mat_elem(b, i, j), mat_elem(b, pivot, j));
        }

        // それでもピボットが0ならば特異行列
        if (fabs(mat_elem(A, i, i)) < epsilon) {
            fprintf(stderr, "[ ERROR ] matrix is singular!\n");
            mat_free(&A);
            mat_free(&b);
            return false;
        }

        // ピボットより下の行を消去する
        for (int j = i + 1; j < A.rows; j++) {
            const double ratio = mat_elem(A, j, i) / mat_elem(A, i, i);
            for (int k = i; k < A.cols; k++) {
                mat_elem(A, j, k) -= ratio * mat_elem(A, i, k);
            }

            for (int k = 0; k < b.cols; k++) {
                mat_elem(b, j, k) -= ratio * mat_elem(b, i, k);
            }
        }
    }

    // 後退代入
    for (int i = A.rows - 1; i >= 0; i--) {
        for (int k = 0; k < b.cols; k++) {
            for (int j = i + 1; j < A.rows; j++) {
                mat_elem(b, i, k) -= mat_elem(A, i, j) * mat_elem(b, j, k);
            }
            mat_elem(b, i, k) /= mat_elem(A, i, i);
        }
    }

    mat_copy(x, b);
    mat_free(&A);
    mat_free(&b);

    return true;
}

// mat_inverse:行列Aの逆行列を*invAに与える
bool mat_inverse(matrix *invA, matrix A)
{
    if (!is_same_size(*invA, A) || A.rows != A.cols) return false;

    matrix temp;
    if (!mat_alloc(&temp, A.rows, A.cols)) return false;
    
    mat_ident(&temp);
    bool ret = mat_solve(invA, A, temp);
    mat_free(&temp);

    return ret;
}
